package speiger.src.coreengine.rendering.gui.font;

import java.util.Map;
import java.util.function.Consumer;

import speiger.src.coreengine.assets.AssetLocation;
import speiger.src.coreengine.rendering.gui.font.glyth.Glyth;
import speiger.src.coreengine.rendering.gui.font.glyth.GlythData;
import speiger.src.coreengine.rendering.gui.font.glyth.UnbakedGlyth.GlythBaker;
import speiger.src.coreengine.rendering.tesselation.buffer.IVertexBuilder;

public class Font {
	private static final int FORMAT_CODE_POINT = '§';
	FontSplitter splitter;
	Map<AssetLocation, FontGroup> fonts;
	GlythBaker baker;
	float oversample;
	GlythCache[] styledCache = new GlythCache[] {new GlythCache(this, 0), new GlythCache(this, 1), new GlythCache(this, 2), new GlythCache(this, 3)};
	
	protected Font(Map<AssetLocation, FontGroup> fonts, GlythBaker baker, float oversample, Consumer<Runnable> clearing) {
		this.fonts = fonts;
		this.baker = baker;
		this.oversample = oversample;
		clearing.accept(this::reset);
		splitter = new FontSplitter(this);
	}
	
	private void reset() {
		for(int i = 0;i<4;i++) {
			styledCache[i].reset();
		}
	}
	
	public GlythData data(FontStyle font, int codepoint) {
		return styledCache[font.style() & 0x3].data(font, codepoint);
	}
	
	public Glyth glyth(FontStyle font, int codepoint) {
		return styledCache[font.style() & 0x3].glyth(font, codepoint);
	}
	
	protected FontGroup font(AssetLocation font) {
		return fonts.get(font);
	}
	
	public float drawText(String text, float x, float y, int color, TexturedBuffer buffer, boolean end) {
		return drawText(TextStyle.DEFAULT, text, x, y, color, buffer, 1F, end);
	}
	
	public float drawText(TextStyle style, String text, float x, float y, int color, TexturedBuffer buffer, float scale, boolean end) {
		TextStyle[] currentStyle = {style};
		int[] currentColor = {currentStyle[0].hasColor() ? currentStyle[0].color() : color};
		float xStart = x;
		int previousCodepoint = -1;
		for(int i = 0,m=text.length();i<m;) {
			int codepoint = text.codePointAt(i);
			if(codepoint == FORMAT_CODE_POINT && applyStyle(text, i+1, currentStyle[0], style, T -> {
				currentStyle[0] = T;
				currentColor[0] = T.hasColor() ? T.color() : color;
			})) {
				i = text.indexOf('>', i)+1;
				continue;
			}
			GlythData data = data(currentStyle[0].font(), codepoint);
			if(previousCodepoint != -1) {
				x -= data.kerning(previousCodepoint);
			}
			Glyth glyth = glyth(currentStyle[0].font(), codepoint);
			if(glyth.isValid()) {
				float minX = (glyth.left() + x) * scale;
				float minY = (glyth.top() + y) * scale;
				float maxX = (glyth.right() + x) * scale;
				float maxY = (glyth.bottom() + y) * scale;
				
				IVertexBuilder builder = buffer.builderForTexture(glyth.texture());
				builder.pos(minX, minY, 0F).tex(glyth.minU(), glyth.minV()).rgba(currentColor[0]).endVertex();
				builder.pos(maxX, minY, 0F).tex(glyth.maxU(), glyth.minV()).rgba(currentColor[0]).endVertex();
				builder.pos(maxX, maxY, 0F).tex(glyth.maxU(), glyth.maxV()).rgba(currentColor[0]).endVertex();
				builder.pos(maxX, maxY, 0F).tex(glyth.maxU(), glyth.maxV()).rgba(currentColor[0]).endVertex();
				builder.pos(minX, maxY, 0F).tex(glyth.minU(), glyth.maxV()).rgba(currentColor[0]).endVertex();
				builder.pos(minX, minY, 0F).tex(glyth.minU(), glyth.minV()).rgba(currentColor[0]).endVertex();
			}
			x += data.advance();
			i += Character.charCount(codepoint);
			previousCodepoint = codepoint;
		}
		if(end) buffer.end();
		return x - xStart;
	}
	
	boolean applyStyle(String text, int index, TextStyle style, TextStyle original, Consumer<TextStyle> consumer) {
		if(text.length() <= index || text.codePointAt(index) != '<') return false;
		int endIndex = text.indexOf('>', index);
		if(endIndex == -1) return false;
		String[] formatting = text.substring(index+1, endIndex).split(",");
		if(formatting.length <= 0) return false;
		consumer.accept(style.parseArguments(original, formatting));
		return true;
	}
	
	
	public static interface TexturedBuffer {
		public IVertexBuilder builderForTexture(int textureId);
		public void end();
	}
}
